/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	DOUBFIFO.C						*/
/*  Version  :	7.9.90/3						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:  Funktionen zur Verwaltung eines Fifo-Speichers.		*/
/*									*/
/*  History  :	Erstversion in Programmprojekt RESYNTH.V40/1988;	*/
/*		10.1.90/1 Umbau fuer Verwendung in ANLYX.V30		*/
/*		9.6.90/1  Aenderung von WORD auf double-Eintraegen.	*/
/*			  Ergaenzung von fillfifo()			*/
/*			  Fehlerbehebung: Fifotiefe = SIZE - 1		*/
/*		7.9.90/3  Ergaenzung von addfifo() und writefifo()	*/
/*									*/
/************************************************************************/

#define	DOUBFIFO_SOURCE			/* Source-Identifikation fuer...*/
#include "doubfifo.h"			/* Import/Export-Datei		*/

/*----------------------------------------------------------------------*/
/*	CREATEFIFO	Installiert und initialisiert einen FIFO der	*/
/*			mit einer effekt. Tiefe von MAXSIZE Elementen.	*/
/*			Die tatsaechliche Feldgroesse ist um einen	*/
/*			Platz groesser wg. Fuellstandsindexverfahren. 	*/
/*			HEAD zeigt auf letzten Der ueber FDP	*/
/*			uebergebene Descriptor wird entsprechend ge-	*/
/*			setzt. Rueckgabewert ungleich 0 bei Misserfolg	*/
/*			(kein Speicher mehr frei).			*/
/*----------------------------------------------------------------------*/
int createfifo(fdp, maxsize)
FIFODESCR *fdp;
int maxsize;
{
	if((fdp->ADR = (FIFOELEM*)lmalloc((long)(maxsize+1) * 
			sizeof(FIFOELEM))) == 0L)
		return(-1);
	fdp->SIZE = maxsize+1;
	clearfifo(fdp);
	return(0);
}
/*----------------------------------------------------------------------*/
/*	CLEARFIFO	Loescht Inhalt des per FDP erreichbaren	Fifos.	*/
/*----------------------------------------------------------------------*/
clearfifo(fdp)
FIFODESCR *fdp;
{
	fdp->HEAD = 0;
	fdp->TAIL = 1;
}
/*----------------------------------------------------------------------*/
/*	FILLFIFO	Fuellt Inhalt des per FDP erreichbaren Fifos	*/
/*			mit Werten ELEM bis an den Kragen voll...	*/
/*----------------------------------------------------------------------*/
fillfifo(fdp, elem)
register FIFODESCR *fdp;
FIFOELEM elem;
{
	while (fifolevel(fdp) < fdp->SIZE - 1) {
		fdp->HEAD++;
		fdp->HEAD %= fdp->SIZE;
		(fdp->ADR)[fdp->HEAD] = elem;
	} 
}
/*----------------------------------------------------------------------*/
/*	DISPOSEFIFO	Loest den ueber FDP erreichbaren FIFO auf 	*/
/*			(Rueckgabe Speicher).				*/
/*----------------------------------------------------------------------*/
disposefifo(fdp)
FIFODESCR *fdp;
{
	clearfifo(fdp);
	free((char*)(fdp->ADR));
	fdp->ADR = 0L;
}
/*----------------------------------------------------------------------*/
/*	FIFOLEVEL	gibt die momentane Fuellung des ueber den FIFO-	*/
/*			Descriptorpointer FDP spezif. FIFOs zurueck.	*/
/*----------------------------------------------------------------------*/
int fifolevel(fdp)
register FIFODESCR *fdp;
{
	return( (fdp->HEAD - fdp->TAIL + fdp->SIZE + 1) % fdp->SIZE );
}
/*----------------------------------------------------------------------*/
/*	ELEMINFIFO	versucht das Element ELEM in FIFO zu schreiben,	*/
/*			Rueckgabewert ungleich null bei Misserfolg.	*/
/*----------------------------------------------------------------------*/
int eleminfifo(fdp, elem)
register FIFODESCR *fdp;
FIFOELEM elem;
{
	if (fifolevel(fdp) < fdp->SIZE - 1) {
		fdp->HEAD++;
		fdp->HEAD %= fdp->SIZE;
		(fdp->ADR)[fdp->HEAD] = elem;
		return(0);
	} else
		return(-1);		/* FIFO ist voll		*/
}
/*----------------------------------------------------------------------*/
/*	OUTELEMFIFO	holt aeltesten Eintrag aus FIFO und uebergibt	*/
/*			ihn ueber ELEMP.				*/
/*			(return) ungleich 0, wenn nicht zugegriffen	*/
/*			werden kann.					*/
/*----------------------------------------------------------------------*/
int outelemfifo(fdp, elemp)
FIFODESCR *fdp;
FIFOELEM *elemp;
{
	if (fifolevel(fdp) <= 0)
		return(-1);
	*elemp = (fdp->ADR)[fdp->TAIL++];
	fdp->TAIL %= fdp->SIZE;
	return(0);
}
/*----------------------------------------------------------------------*/
/*	WRITEFIFO	ueberschreibt Eintrag in FIFO mit dem durch	*/
/*			ELEM uebergebenen Wert; INDEX = -1 fordert	*/
/*			letzten Eintrag an, -2 den vorletzten, etc.	*/
/*			(return) ungleich 0, wenn nicht zugegriffen	*/
/*			werden kann.					*/
/*----------------------------------------------------------------------*/
int writefifo(fdp, index, elem)
register FIFODESCR *fdp;
register int index;
FIFOELEM elem;
{
	if (fifolevel(fdp) >= -index) {
	  if (index < 0) {
	    (fdp->ADR)[(fdp->HEAD + fdp->SIZE + ++index) % fdp->SIZE] =
	      elem;
	    return(0);
	  }
	}
	return(-1);
}
/*----------------------------------------------------------------------*/
/*	ADDFIFO		addiert zu Eintrag in FIFO den durch		*/
/*			ELEM uebergebenen Wert; INDEX = -1 fordert	*/
/*			letzten Eintrag an, -2 den vorletzten, etc.	*/
/*			(return) ungleich 0, wenn nicht zugegriffen	*/
/*			werden kann.					*/
/*----------------------------------------------------------------------*/
int addfifo(fdp, index, elem)
register FIFODESCR *fdp;
register int index;
FIFOELEM elem;
{
	if (fifolevel(fdp) >= -index) {
	  if (index < 0) {
	    (fdp->ADR)[(fdp->HEAD + fdp->SIZE + ++index) % fdp->SIZE] +=
	      elem;
	    return(0);
	  }
	}
	return(-1);
}
/*----------------------------------------------------------------------*/
/*	READFIFO	liest die letzen Eintraege aus FIFO; INDEX = -1	*/
/*			fordert letzten Eintrag an, -2 den vorletzten,	*/
/*			etc. und uebergibt ihn ueber ELEMP.		*/
/*			(return) ungleich 0, wenn nicht zugegriffen	*/
/*			werden kann.					*/
/*----------------------------------------------------------------------*/
int readfifo(fdp, index, elemp)
register FIFODESCR *fdp;
register int index;
FIFOELEM *elemp;
{
	if (fifolevel(fdp) >= -index) {
	  if (index < 0) {
	    *elemp =
	      (fdp->ADR)[(fdp->HEAD + fdp->SIZE + ++index) % fdp->SIZE];
	    return(0);
	  }
	}
	return(-1);
}
/*******Dateiende********************************************************/
