/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	FILTFLP.C						*/
/*  Version  :	8.8.98/10						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:  Koeffizientenbestimmung und Rechnung von verschiedenen	*/
/*		digitalen rekursiven Tiefpass- und Bandpassfiltern.	*/
/*		Zum Verstaendnis bzw. zur Ergaenzung der moeglichen	*/
/*		Filter folge SCHRITT 1 - 4. Begleitende Literatur:	*/
/*									*/
/*		    M.Herpy,J.-C.Berka,"Aktive RC-Filter",Franzis'	*/
/*		    W.Hess,"Digitale Filter",Teubner (kompliziert)	*/
/*		    U.Tietze, C.Schenk, "Halbleiter-Schaltungtechnik"	*/
/*									*/
/*  History  :	23.3.90/1-3     Erstserie				*/
/*		16.5.90/4	Sektionskoeffizienten durch Makro;	*/
/*				Filter B84N (Nr.9) entfernt.		*/
/*		17.5.90/5	Bandpasseinbau				*/
/*		22.5.90/6	Verschoenerung				*/
/*		25.7.90/7	Makros durch Funktionen ersetzt		*/
/*		13.5.91/8	Bandpasstransformation Faktor 2 zugef.	*/
/*				Filter WL801 (Nr.6) entfernt, Filter	*/
/*				P8, P3, PG3, B3 (Nr.16-19) zugefuegt.	*/
/*				Fehler in filterparms B4(BP) beseitigt.	*/
/*		4.8.94/9	Drastische Vereinfachung der zugrunde-	*/
/*				liegenden Rechung durch direkte		*/
/*				Bilineartransformation.			*/
/*				Berichtigung von F3_ALPHA1_8 effektiv	*/
/*				in der letzten Stelle.			*/
/*		8.8.98/10	PI und ZPI defs nur wenn undefiniert	*/
/*									*/
/************************************************************************/

#define	FILTFLP_SOURCE			/* Source-Identifikation fuer...*/
#include "filtflp.h"			/* Import/Export-Datei		*/

/*----------------------------------------------------------------------*/
/*  SCHRITT 1 : globale Filtertypentabelle				*/
/*  ======================================				*/
/*									*/
/*	Das Hinzufuegen eines neuen Filters beginnt hier; die Tabelle	*/
/*	hat 3 Eintraege, der erste ist die Filterkennung, der fuer die	*/
/*	Auswahl der Rechnungen entscheidend ist. Zweiter Eintrag stellt	*/
/*	Ein Kuerzel dar, der dritte die volle Bezeichnung.		*/
/*----------------------------------------------------------------------*/

FTYPTABENTRY g_ftyptab[] = {
	1, "P1",     "Grad 1",
	2, "2P1",    "2x Grad 1",
	14,"3P1",    "3x Grad 1",
	15,"4P1",    "4x Grad 1",
	3, "8P1",    "8x Grad 1",
	4, "B2",     "Bessel Grad 2",
	17,"B3",     "Bessel Grad 3",
	8, "B4",     "Bessel Grad 4",
	5, "B8",     "Bessel Grad 8",
	11,"PG2",    "Pseudo-Gauss Grad 2",
	19,"PG3",    "Pseudo-Gauss Grad 3",
	13,"PG4",    "Pseudo-Gauss Grad 4",
	10,"PG8",    "Pseudo-Gauss Grad 8",
	12,"P2",     "Potenz Grad 2",
	18,"P3",     "Potenz Grad 3",
	7, "P4",     "Potenz Grad 4",
	16,"P8",     "Potenz Grad 8",
	0, "\0", "\0"
};
/*----------------------------------------------------------------------*/
/*  SCHRITT 2 : Filterdaten						*/
/*  =======================						*/
/*									*/
/*	Die Pole und Konstanten der normierten Uebertragungsfunktion	*/
/*	H(w) (H(0)==1; w3dB==1) des analogen Tiefpasses werden hier	*/
/*	definiert, die Ziffer im Namens-Prefix sollte mit der Filter-	*/
/*	kennung (s.o) uebereinstimmen.					*/
/*----------------------------------------------------------------------*/
					/* Real/+-Imaginaerteil Pole	*/
					/* und Daempfungskonstante	*/
					/* normiert auf 3dB Grenzk.freq.*/	
					/* und Daempfung 0 bei w=0	*/
					/* (CEH im Nenner der UeF !)	*/

#ifndef PI
#define PI		3.141592654	/* PI				*/
#endif
#ifndef ZPI
#define ZPI		(2*PI)
#endif

#define F1_ALPHA1      -1.0		/* 1xTP 1. Ordnung		*/
#define F1_CEH		1.0
#define F1_TG_AT0	1.0

#define F2_ALPHA1_2    -1.55377		/* 2xTP 1. Ordnung		*/
#define F2_CEH		0.41421
#define F2_TG_AT0	1.2872

#define F3_ALPHA1_8    -3.32397		/* 8xTP 1. Ordnung		*/
#define F3_CEH		0.000067098
#define F3_TG_AT0	2.4067

#define F4_ALPHA1      -1.10160		/* Thomson (Bessel) TP Grad 2	*/
#define F4_BETA1	0.63601
#define F4_CEH		0.61803		
#define F4_TG_AT0	1.3617

#define F5_ALPHA1      -1.75741		/* Thomson (Bessel) TP Grad 8	*/
#define F5_ALPHA2      -1.63694
#define F5_ALPHA3      -1.37384
#define F5_ALPHA4      -0.89287
#define F5_BETA1	0.27287
#define F5_BETA2	0.82280
#define F5_BETA3	1.38836
#define F5_BETA4	1.99833
#define F5_CEH		0.0051541
#define F5_TG_AT0	3.1796

#define F7_ALPHA1      -0.92388		/* Potenz Tiefpass Grad 4	*/
#define F7_ALPHA2      -0.38268
#define F7_BETA1	F7_ALPHA2
#define F7_BETA2	F7_ALPHA1
#define F7_CEH		1.0
#define F7_TG_AT0	2.6131

#define F8_ALPHA1      -1.37001		/* Thomson (Bessel) TP Grad 4	*/
#define F8_ALPHA2      -0.99521
#define F8_BETA1	0.41025
#define F8_BETA2	1.25711
#define F8_CEH		0.19019
#define F8_TG_AT0	2.1140

#define F10_ALPHA1     -1.49815		/* Pseudo-Gaussfilter Grad 8	*/
#define F10_ALPHA2     -1.83307		/* aus:				*/
#define F10_ALPHA3     -2.01063		/*  Milton Dishal:		*/
#define F10_ALPHA4     -2.09049		/*  Gaussian-Response Filter 	*/
#define F10_BETA1 	2.27978		/*  Design, Electrical Com-	*/
#define F10_BETA2	1.48938		/*  munication,Vol.36,No.1,1959	*/
#define F10_BETA3	0.85597
#define F10_BETA4	0.28007
#define F10_CEH		0.00113393
#define F10_TG_AT0	2.8418

#define F11_ALPHA1     -1.28963		/* Pseudo-Gaussfilter Grad 2	*/
#define F11_BETA1	0.53418
#define F11_CEH		0.51321		
#define F11_TG_AT0	1.3237

#define F12_ALPHA1     -0.70710		/* Potenz-TP Grad 2		*/
#define F12_BETA1	F12_ALPHA1
#define F12_CEH		1.0
#define F12_TG_AT0	1.4142

#define F13_ALPHA1     -1.42106		/* Pseudo-Gaussfilter Grad 4	*/
#define F13_ALPHA2     -1.63074
#define F13_BETA1	1.27582
#define F13_BETA2	0.39458
#define F13_CEH		0.097402
#define F13_TG_AT0	1.9379

#define F14_ALPHA1_3   -1.96146		/* 3xTP 1.Ordnung		*/
#define F14_CEH		0.13251
#define F14_TG_AT0	1.5295

#define F15_ALPHA1_4   -2.29896		/* 4xTP 1.Ordnung		*/
#define F15_CEH		0.035799
#define F15_TG_AT0	1.7399

#define F16_ALPHA1     -0.98079		/* Potenz-TP Grad 8		*/
#define F16_ALPHA2     -0.83147
#define F16_ALPHA3     -0.55557
#define F16_ALPHA4     -0.19509
#define F16_BETA1	F16_ALPHA4
#define F16_BETA2	F16_ALPHA3
#define F16_BETA3	F16_ALPHA2
#define F16_BETA4	F16_ALPHA1
#define F16_CEH		1.00000
#define F16_TG_AT0	5.1258

#define F17_ALPHA1     -1.04741		/* Thomson (Bessel) TP Grad 3	*/
#define F17_ALPHA2     -1.32268
#define F17_BETA1	0.99926
#define F17_CEH		0.36078
#define F17_TG_AT0	1.7557

#define F18_ALPHA1     -0.50000		/* Potenz-TP Grad 3		*/
#define F18_ALPHA2     -1.00000
#define F18_BETA1	0.86603
#define F18_CEH		1.00000
#define F18_TG_AT0	1.4142

#define F19_ALPHA1     -1.38257		/* Pseudo-Gaussfilter Grad 3	*/
#define F19_ALPHA2     -1.52005
#define F19_BETA1	0.94621
#define F19_CEH		0.23438
#define F19_TG_AT0	1.6430

/*----------------------------------------------------------------------*/
/*  BILINTRANS1/2							*/
/*									*/
/*		Bilineartransformation f. Nenner- und Zaehlerfaktoren	*/
/*		1. bzw. 2. Grades der Uebertragungsfunktion. Uebergang	*/
/*		von normierte P-Ebene (normiert auf w3dB) in z-Ebene 	*/
/*		durch P=k*(z-1)/(z+1).					*/
/*									*/
/*									*/
/*   Grad 1: (P-alpha) 		-> (s0*z+s1)/(z+1)			*/
/*   Grad 2: (P-2*alpha*P+(alpha**2+beta**2))				*/
/*				-> (s0*z**2+s1*z+s2)/(z+1)**2		*/
/*									*/
/*----------------------------------------------------------------------*/
static bilintrans1(alpha,s0p,s1p,k)
double alpha, *s0p, *s1p, k;
{
    *s0p = k - alpha;
    *s1p =-k - alpha;
}

static bilintrans2(alpha,beta,s0p,s1p,s2p,k)
double alpha, beta, *s0p, *s1p, *s2p, k;
{
    *s0p = (k-alpha)*(k-alpha) + beta*beta;
    *s1p = 2 * (alpha*alpha + beta*beta - k*k);
    *s2p = (k+alpha)*(k+alpha) + beta*beta;
}
/*----------------------------------------------------------------------*/
/*  FILTCOEF1/2P							*/
/*									*/
/*		berechnet Koeffizienten der Filterdifferenzengleichung	*/
/*		fuer Teiluebertragungsfunktionen, so, wie fuer Rechen-	*/
/*		funktionen FILTER1/2P benoetigt.			*/
/*									*/
/*   Grad 1:	(z+1)/(s0*z+s1) -> y+g1*y_1 = c*(x+x_1)			*/
/*   Grad 2:	(z+1)**2/(s0*z**2+s1*z+s2) ->				*/
/* 			  y+g1*y_1+g2*y_2 =  c*(x+2*x_1+x_2)		*/
/*----------------------------------------------------------------------*/
static filtcoef1p(s0,s1,cp,g1p)
double s0, s1, *cp, *g1p;
{
    *g1p = s1 / s0;
    *cp  = 1. / s0;
}

static filtcoef2p(s0,s1,s2,cp,g1p,g2p)
double s0, s1, s2, *cp, *g1p, *g2p;
{
    *g1p = s1 / s0;
    *g2p = s2 / s0;
    *cp  = 1. / s0;
}
/*----------------------------------------------------------------------*/
/*  BP1N2P_FROM_1P bzw. BP2N4P_FROM_2P	(nur fuer Bandpaesse)		*/
/*									*/
/*		ueberfuehrt den aus FILTCOEF1P/2P gewonnenen		*/
/*		Filterkoeffizientensatz einer TP-Filtersektion 1. bzw.	*/
/*		2. Grades in einen fuer BP-Filtersektion mit Mitten-	*/
/*		frequenz fm. Koeffizientensatz so, wie fuer BANDPASS-	*/
/*		1N2P/2N4P-Rechnefunktion benoetigt.			*/
/*									*/
/* 	Die Umwandlung von Tiefpaessen in Bandpaesse erfolgt durch 2	*/
/*	spektral gegensinnige Rotationen mit Ueberlagerung nach fol-	*/
/*	gendem Schema:							*/
/*									*/
/*	    HBP(z) = (HTP(z*exp(A)) + HTP(z*exp(-A))) * cos(phi) -	*/
/*		 j * (HTP(z*exp(A)) - HTP(z*exp(-A))) * sin(phi)	*/
/*									*/
/*	    hBP(n) = hTP(n) * 2 * cos(A*n + phi)			*/
/*									*/
/*	    A=j*2*pi*fm*T, Abtastperiode T, Mittenfrequenz fm		*/
/*									*/
/*	Dieses Prinzip ist i.F. auf Teiltuebertragungsfunktionen	*/
/*	1. und 2. Grades einzeln anzuwenden, was eine zulaessige Naehe-	*/
/*	rung im Falle ausreichender Bandfilterguete darstellt. Dazu 	*/
/*	erfolgt eine Transformation der Nennerkoef. einer TP-Teil-UeF	*/
/*	1. oder	2. Grades in Koeffizienten einer Bandpass-Teil-UeF	*/
/*	2. bzw. 4, Grades. Die von der Bilineartransformation stammen-	*/
/*	den Nullstellenfaktoren (z+1) in der TP-Teil-Uef werden nicht	*/
/*	mittransformiert, sondern jeweils durch den konstanten Faktor 2	*/
/*	angenaehert.							*/
/*----------------------------------------------------------------------*/
static bp1n2p_from_1p(g1,fm,fa,ph,d1ap,d2ap,g1ap,g2ap)
double g1, fm, fa, ph, *d1ap, *d2ap, *g1ap, *g2ap;
{
	*d1ap = 4*cos(ZPI*fm/fa+ph);	/* d0a = 0 			*/
	*d2ap = 4*g1*cos(ph);
	*g1ap= g1*2*cos(ZPI*fm/fa);
	*g2ap= g1*g1;
}

static bp2n4p_from_2p(g1,g2,fm,fa,ph,d0ap,d1ap,d2ap,g1ap,g2ap,g1bp,g2bp)
double g1, g2, fm, fa, ph, *d0ap, *d1ap, *d2ap, *g1ap, *g2ap, *g1bp, *g2bp;
{
	*d0ap = 8*cos(2*ZPI*fm/fa+ph);	/* d0b = 0, d1b = 0, d2b = 1	*/
	*d1ap = 8*g1*cos(ZPI*fm/fa+ph);
	*d2ap = 8*g2*cos(ph);
	*g1ap = g1*cos(ZPI*fm/fa)+
			sqrt(4*g2-g1*g1)*sin(ZPI*fm/fa);
	*g2ap = g2;
	*g1bp = g1*cos(ZPI*fm/fa)-	
			sqrt(4*g2-g1*g1)*sin(ZPI*fm/fa);
	*g2bp = g2;
}
/*----------------------------------------------------------------------*/
/*  SCHRITT 3 : FILTERPARMS zur Berechnung Filterkoeffizienten 		*/
/*  ==========================================================		*/
/*									*/
/*    ->|	code 	Filterkennung,falls neg.Vorzeichen -> Bandpass	*/
/*			hBP(t)=hTP(t)*cos(2*pi*fm*t+ph)			*/
/*		fa	Abtastfrequenz in Hz				*/
/*		f3dB	3dB-Grenzfrequenz				*/
/*		fm      Mittenfrequenz in Hz, falls Bandpass		*/
/*		ph	Phase fuer Bandpasstransformation		*/
/*	|->	d,g	Koeffizienten-Arrays				*/
/*									*/
/*	Fuer jedes Filter muss eine eigne Koeffizientenberechnungs-	*/
/*	routine im Rahmen der der durch CODE gesteuerten switch()-	*/
/*	Kontrollstruktur eingetragen werden. Man verfolge hierzu	*/
/*	die exemplarische Kommentierung zum 1. Filter.			*/
/*									*/
/*  ACHTUNG! 	keine Koeffizienten g,d mit Index >= MAXG bzw. MAXD	*/
/*	 	(definiert in filtflp.h) verwenden. Das rufende Pro-	*/
/*		gramm muss die entsprechenden Feldgroessen MAXG/MAXD	*/
/*		bereitstellen.						*/
/*----------------------------------------------------------------------*/
double filterparms(code, fa, f3db, fm, ph, d, g)
int code;
double fa, f3db, fm, ph, d[], g[];
{
	double	
	    sa0,sa1,sa2,sb0,sb1,sb2,sc0,sc1,sc2,
	    sd0,sd1,sd2,se0,se1,se2,sf0,sf1,sf2,
	    ga1,ga2,gb1,gb2,gc1,gc2,
	    gd1,gd2,ge1,ge2,gf1,gf2,
	    da0,da1,da2,db0,db1,db2,dc0,dc1,dc2,
	    dd0,dd1,dd2,de0,de1,de2,df0,df1,df2,
	    ca,cb,cc,cd,ce,cf,cges,
	    k,tg_at0;

	k = 1./tan(PI * f3db / fa);	/* Entnormierungsfunktion, s.u.	*/
	
	switch(abs(code)) {
	case 1:

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
					/* F1 - Rechnung exemplarisch	*/
					/*				*/
					/* Transformation der Wurzel-	*/
					/* faktoren 1. und 2.Grades und	*/
					/* Ue-Konstante (Grad 0) der	*/
					/* norm. analog. UeF H(P) in	*/
					/* diskrete TeilUeFs Ht(z) ..	*/
					/* H0(P)=c -> H0(z)=c		*/
					/* H1(P)=(P-a) ->		*/
					/* H1(z)=(s0*z+s1)/(z+1)	*/
					/* H2(P)=(P**2-2*a*P+a**2+b**2)	*/
					/* H2(z)=(s0*z**2+s1*z+s2)/	*/
					/*       (z+1)**2		*/
					/* ...oder Kehrwerte		*/
		bilintrans1(F1_ALPHA1,&sa0,&sa1,k);

					/* Zusammenfassung aller Ht(z)	*/
					/* zu H(z),ggf. Kuerzen (zB z+1)*/
					/* und Aufspaltung H(z) in Teil-*/
					/* sektionen vom Grad 0,1,2.	*/
					/* Wahl Teilsektion FILTERXX in	*/
					/* filtflp.h und Bestimmung der	*/
					/* Koeffizienten durch geeigne-	*/
					/* te Normierung per FILTCOEFXX.*/
					/* Beruecksichtung der Konstan-	*/
					/* ten aller Teilsektionen in	*/
					/* einer "Verstaerkersektion"	*/
					/* AMPLIFIER mit Verstaerkung	*/
					/* cges.			*/
		filtcoef1p(sa0,sa1,&ca,&ga1);
		cges = ca / F1_CEH;
					/* Abspeichern in Koeffizien-	*/
					/* Array; Zugriff hierauf muss	*/
					/* mit FILTERING in filtflp.h	*/
					/* korrespondieren.		*/
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;	/* ggf. Bandpasstransformation	*/
		} else {
			bp1n2p_from_1p(ga1,fm,fa,ph,
				d+1,d+2,g+0,g+1);
		}
					/* norm. Gruppenlaufzeit	*/
		tg_at0 = F1_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 2:				/* F2				*/
		bilintrans1(F2_ALPHA1_2,&sa0,&sa1,k);
		filtcoef1p(sa0,sa1,&ca,&ga1);
		cges = ca * ca / F2_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
		} else {
			bp1n2p_from_1p(ga1,fm,fa,ph,
				d+1,d+2,g+0,g+1);
		}
		tg_at0 = F2_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 3:				/* F3				*/
		bilintrans1(F3_ALPHA1_8,&sa0,&sa1,k);
		filtcoef1p(sa0,sa1,&ca,&ga1);
		cges = ca*ca*ca*ca*ca*ca*ca*ca / F3_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
		} else {
			bp1n2p_from_1p(ga1,fm,fa,ph,
				d+1,d+2,g+0,g+1);
		}
		tg_at0 = F3_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 4:				/* F4				*/
		bilintrans2(F4_ALPHA1,F4_BETA1,&sa0,&sa1,&sa2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		cges = ca / F4_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
		}
		tg_at0 = F4_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 5:				/* F5				*/
		bilintrans2(F5_ALPHA1,F5_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F5_ALPHA2,F5_BETA2,&sb0,&sb1,&sb2,k);
		bilintrans2(F5_ALPHA3,F5_BETA3,&sc0,&sc1,&sc2,k);
		bilintrans2(F5_ALPHA4,F5_BETA4,&sd0,&sd1,&sd2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		filtcoef2p(sc0,sc1,sc2,&cc,&gc1,&gc2);
		filtcoef2p(sd0,sd1,sd2,&cd,&gd1,&gd2);
		cges = ca * cb * cc * cd / F5_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
			g[4] = gc1;
			g[5] = gc2;
			g[6] = gd1;
			g[7] = gd2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
			bp2n4p_from_2p(gc1,gc2,fm,fa,ph,
				d+7,d+8,d+9,g+8,g+9,g+10,g+11);
			bp2n4p_from_2p(gd1,gd2,fm,fa,ph,
				d+10,d+11,d+12,
				g+12,g+13,g+14,g+15);
		}
		tg_at0 = F5_TG_AT0;
		break;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 7:				/* F7				*/
		bilintrans2(F7_ALPHA1,F7_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F7_ALPHA2,F7_BETA2,&sb0,&sb1,&sb2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		cges = ca * cb / F7_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
		}
		tg_at0 = F7_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 8:				/* F8				*/
		bilintrans2(F8_ALPHA1,F8_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F8_ALPHA2,F8_BETA2,&sb0,&sb1,&sb2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		cges = ca * cb / F8_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
		}
		tg_at0 = F8_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 10:			/* F10				*/
		bilintrans2(F10_ALPHA1,F10_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F10_ALPHA2,F10_BETA2,&sb0,&sb1,&sb2,k);
		bilintrans2(F10_ALPHA3,F10_BETA3,&sc0,&sc1,&sc2,k);
		bilintrans2(F10_ALPHA4,F10_BETA4,&sd0,&sd1,&sd2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		filtcoef2p(sc0,sc1,sc2,&cc,&gc1,&gc2);
		filtcoef2p(sd0,sd1,sd2,&cd,&gd1,&gd2);
		cges = ca * cb * cc * cd / F10_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
			g[4] = gc1;
			g[5] = gc2;
			g[6] = gd1;
			g[7] = gd2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
			bp2n4p_from_2p(gc1,gc2,fm,fa,ph,
				d+7,d+8,d+9,g+8,g+9,g+10,g+11);
			bp2n4p_from_2p(gd1,gd2,fm,fa,ph,
				d+10,d+11,d+12,
				g+12,g+13,g+14,g+15);
		}
		tg_at0 = F10_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 11:			/* F11				*/
		bilintrans2(F11_ALPHA1,F11_BETA1,&sa0,&sa1,&sa2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		cges = ca / F11_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
		}
		tg_at0 = F11_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 12:			/* F12				*/
		bilintrans2(F12_ALPHA1,F12_BETA1,&sa0,&sa1,&sa2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		cges = ca / F12_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
		}
		tg_at0 = F12_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
	case 13:			/* F13				*/
		bilintrans2(F13_ALPHA1,F13_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F13_ALPHA2,F13_BETA2,&sb0,&sb1,&sb2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		cges = ca * cb / F13_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
		}
		tg_at0 = F13_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 14:			/* F14				*/
		bilintrans1(F14_ALPHA1_3,&sa0,&sa1,k);
		filtcoef1p(sa0,sa1,&ca,&ga1);
		cges = ca * ca * ca / F14_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
		} else {
			bp1n2p_from_1p(ga1,fm,fa,ph,
				d+1,d+2,g+0,g+1);
		}
		tg_at0 = F14_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 15:			/* F15				*/
		bilintrans1(F15_ALPHA1_4,&sa0,&sa1,k);
		filtcoef1p(sa0,sa1,&ca,&ga1);
		cges = ca * ca * ca * ca / F15_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
		} else {
			bp1n2p_from_1p(ga1,fm,fa,ph,
				d+1,d+2,g+0,g+1);
		}
		tg_at0 = F15_TG_AT0;
		break;

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	case 16:			/* F16				*/
		bilintrans2(F16_ALPHA1,F16_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans2(F16_ALPHA2,F16_BETA2,&sb0,&sb1,&sb2,k);
		bilintrans2(F16_ALPHA3,F16_BETA3,&sc0,&sc1,&sc2,k);
		bilintrans2(F16_ALPHA4,F16_BETA4,&sd0,&sd1,&sd2,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef2p(sb0,sb1,sb2,&cb,&gb1,&gb2);
		filtcoef2p(sc0,sc1,sc2,&cc,&gc1,&gc2);
		filtcoef2p(sd0,sd1,sd2,&cd,&gd1,&gd2);
		cges = ca * cb * cc * cd / F16_CEH,
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
			g[3] = gb2;
			g[4] = gc1;
			g[5] = gc2;
			g[6] = gd1;
			g[7] = gd2;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp2n4p_from_2p(gb1,gb2,fm,fa,ph,
				d+4,d+5,d+6,g+4,g+5,g+6,g+7);
			bp2n4p_from_2p(gc1,gc2,fm,fa,ph,
				d+7,d+8,d+9,g+8,g+9,g+10,g+11);
			bp2n4p_from_2p(gd1,gd2,fm,fa,ph,
				d+10,d+11,d+12,
				g+12,g+13,g+14,g+15);
		}
		tg_at0 = F16_TG_AT0;
		break;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 17:				/* F17			*/
		bilintrans2(F17_ALPHA1,F17_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans1(F17_ALPHA2,&sb0,&sb1,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef1p(sb0,sb1,&cb,&gb1);
		cges = ca * cb / F17_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp1n2p_from_1p(gb1,fm,fa,ph,
				d+4,d+5,g+4,g+5);
		}
		tg_at0 = F17_TG_AT0;
		break;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 18:				/* F18			*/
		bilintrans2(F18_ALPHA1,F18_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans1(F18_ALPHA2,&sb0,&sb1,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef1p(sb0,sb1,&cb,&gb1);
		cges = ca * cb / F18_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp1n2p_from_1p(gb1,fm,fa,ph,
				d+4,d+5,g+4,g+5);
		}
		tg_at0 = F18_TG_AT0;
		break;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
 	case 19:				/* F19			*/
		bilintrans2(F19_ALPHA1,F19_BETA1,&sa0,&sa1,&sa2,k);
		bilintrans1(F19_ALPHA2,&sb0,&sb1,k);
		filtcoef2p(sa0,sa1,sa2,&ca,&ga1,&ga2);
		filtcoef1p(sb0,sb1,&cb,&gb1);
		cges = ca * cb / F19_CEH;
		d[0] = cges;
		if (code > 0) {
			g[0] = ga1;
			g[1] = ga2;
			g[2] = gb1;
		} else {
			bp2n4p_from_2p(ga1,ga2,fm,fa,ph,
				d+1,d+2,d+3,g+0,g+1,g+2,g+3);
			bp1n2p_from_1p(gb1,fm,fa,ph,
				d+4,d+5,g+4,g+5);
		}
		tg_at0 = F19_TG_AT0;
		break;


/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
	}				/* Laufzeit in Filtermitte in s	*/
	return(tg_at0 / f3db / ZPI);
}
/*----------------------------------------------------------------------*/
/*  TEILFILTERMAKROS fuer Funktion FILTERING				*/
/*									*/
/*	Die Teilfilter (== Teiluebertragungsfunktionen) werden durch 	*/
/*	Aufbau in 1. kanonischer Form nach folgender Terminologie der	*/
/*	z-Uebertragungsfunktion/Differenzengleichung realisiert:	*/
/*									*/
/*	    AUS          d0 + d1*z**(-1) + d2*z**(-2) + ...		*/	
/*	   -----   =   ---------------------------------------		*/
/*	    EIN          1  + g1*z**(-1) + g2*z**(-2) + ...		*/
/*									*/
/*  aus + g1*aus_1 + g2*aus_2 + ... = d0*ein + d1*ein_1 + d2*ein_2 + ..	*/
/*									*/
/*  	Dabei ist mit _i der um i Abtastwerte verzoegerte Wert gemeint,	*/
/*	mit EIN und AUS die Z-Transformierten von ein bzw. aus.		*/
/*	Die zn-Argumente sind die Zustandsspeicher.			*/
/*									*/
/*	Fuer die Teilfilter FILTER1P,FILTER2P sind die di bereits so	*/
/*	eingestellt, dass sich die Zaehlerfaktoren (z**-1+1) ergeben.	*/
/*----------------------------------------------------------------------*/

#define AMPLIFIER(ein,aus,d0) {		(aus) = (d0) * (ein);		\
					(ein) =        (aus);		\
}
#define FILTER1P(ein,aus,z1,g1)  {  	(aus) =        (ein)   		\
					      +        (z1) ;  		\
					(z1)  =        (ein)   		\
					      - (g1) * (aus);  		\
					(ein) =        (aus);		\
}
#define FILTER2P(ein,aus,z1,z2,g1,g2) {	(aus) =        (ein)   		\
					      +        (z1);   		\
					(z1)  =    2 * (ein)   		\
					      +        (z2)    		\
					      - (g1) * (aus);  		\
					(z2)  =        (ein)   		\
					      - (g2) * (aus);  		\
					(ein) =        (aus);		\
}
#define FILT1N1P(ein,aus,z1,g1,d1)  {  	(aus) =        (ein)   		\
					      +        (z1) ;  		\
					(z1)  = (d1) * (ein)   		\
					      - (g1) * (aus);  		\
					(ein) =        (aus);		\
}
#define FILTER2N(ein,aus,z1,z2,d1,d2) {	(aus) =        (ein)   		\
					      +        (z1);   		\
					(z1)  = (d1) * (ein)   		\
					      +        (z2);   		\
					(z2)  = (d2) * (ein);  		\
					(ein) =        (aus);		\
}
#define FILT2N2P(ein,aus,z1,z2,d1,d2,g1,g2) { 				\
					(aus) =        (ein)   		\
					      +        (z1);   		\
					(z1)  = (d1) * (ein)   		\
					      - (g1) * (aus)   		\
					      +        (z2);   		\
					(z2)  = (d2) * (ein)   		\
					      - (g2) * (aus);  		\
					(ein) =        (aus);		\
}
#define BANDPASS1N2P(ein,aus,z1,z2,d0,d1,g1,g2) {			\
					(aus) = (d0) * (ein)   		\
					      +        (z1);   		\
					(z1)  = (d1) * (ein)   		\
					      - (g1) * (aus)   		\
					      +        (z2);   		\
					(z2)  =                		\
					      - (g2) * (aus);  		\
					(ein) =        (aus);		\
}
#define BANDPASS2N4P(ein,aus,z1,z2,z3,z4,d0,d1,d2,gm1,gm2,gn1,gn2) {	\
									\
					(aus) = (d0) * (ein)   		\
					      +        (z1);   		\
					(z1)  = (d1) * (ein)   		\
					      - (gm1)* (aus)   		\
					      +        (z2);   		\
					(z2)  = (d2) * (ein)   		\
					      - (gm2)* (aus);  		\
					(ein) =        (aus);		\
									\
					(aus) =        (ein)   		\
					      +        (z3);   		\
					(z3)  =                		\
					      - (gn1)* (aus)   		\
					      +        (z4);   		\
					(z4)  =                		\
					      - (gn2)* (aus);  		\
					(ein) =        (aus);		\
}
/*----------------------------------------------------------------------*/
/*  SCHRITT 4 :	FILTERING rechnet Filter um einen Takt weiter.		*/
/*  ==========================================================		*/
/*									*/
/*    ->|	code 	Filterkennung,falls neg.Vorzeichen -> Bandpass	*/
/*		d,g	Koeffizienten-Arrays  (aus FILTERPARMS)		*/
/*		z	Zustandsspeicherarray				*/
/*		ein	Eingangsabtastwert				*/
/*		aus	Ausgangsabtastwert				*/
/*									*/
/*	Fuer jedes Filter muss eine eigne Rechnenroutine im Rahmen der 	*/
/*	durch CODE gesteuerten switch()-Kontrollstruktur eingetragen	*/
/*	Dabei kommen die oben definierten Makros zum Einsatz.		*/
/*----------------------------------------------------------------------*/
filtering(code,ein,z,d,g,ausp)
int code;
double ein, *z, *d, *g, *ausp;
{
	double aus;

	switch((code) > 0 ? (code) : -(code)) {
	case 1:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER1P((ein),(aus),(z)[0],(g)[0]);
		} else {
			BANDPASS1N2P((ein),(aus),(z)[0],(z)[1],
				(d)[1],(d)[2],(g)[0],(g)[1]);
		}
		break;
	case 2:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER1P((ein),(aus),(z)[0],(g)[0]);
			FILTER1P((ein),(aus),(z)[1],(g)[0]);
		} else {
			BANDPASS1N2P((ein),(aus),(z)[0],(z)[1],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[2],(z)[3],
				(d)[1],(d)[2],(g)[0],(g)[1]);
		}
		break;
	case 4:
	case 11:
	case 12:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER2P((ein),(aus),(z)[0],(z)[1],(g)[0],(g)[1]);
		} else {
			BANDPASS2N4P((ein),(aus),
				(z)[0],(z)[1],(z)[2],(z)[3],
				(d)[1],(d)[2],(d)[3],
				(g)[0],(g)[1],(g)[2],(g)[3]);
		}
		break;
	case 17:
	case 18:
	case 19:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER2P((ein),(aus),(z)[0],(z)[1],(g)[0],(g)[1]);
			FILTER1P((ein),(aus),(z)[2],(g)[2]);
		} else {
			BANDPASS2N4P((ein),(aus),
				(z)[0],(z)[1],(z)[2],(z)[3],
				(d)[1],(d)[2],(d)[3],
				(g)[0],(g)[1],(g)[2],(g)[3]);
			BANDPASS1N2P((ein),(aus),(z)[4],(z)[5],
				(d)[4],(d)[5],(g)[4],(g)[5]);
		}
		break;
	case 14:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER1P((ein),(aus),(z)[0],(g)[0]);
			FILTER1P((ein),(aus),(z)[1],(g)[0]);
			FILTER1P((ein),(aus),(z)[2],(g)[0]);
		} else {
			BANDPASS1N2P((ein),(aus),(z)[0],(z)[1],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[2],(z)[3],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[4],(z)[5],
				(d)[1],(d)[2],(g)[0],(g)[1]);
		}
		break;
	case 15:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER1P((ein),(aus),(z)[0],(g)[0]);
			FILTER1P((ein),(aus),(z)[1],(g)[0]);
			FILTER1P((ein),(aus),(z)[2],(g)[0]);
			FILTER1P((ein),(aus),(z)[3],(g)[0]);
		} else {
			BANDPASS1N2P((ein),(aus),(z)[0],(z)[1],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[2],(z)[3],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[4],(z)[5],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[6],(z)[7],
				(d)[1],(d)[2],(g)[0],(g)[1]);
		}
		break;
	case 7:
	case 8:
	case 13:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER2P((ein),(aus),(z)[0],(z)[1],(g)[0],(g)[1]);
			FILTER2P((ein),(aus),(z)[2],(z)[3],(g)[2],(g)[3]);
		} else {
			BANDPASS2N4P((ein),(aus),
				(z)[0],(z)[1],(z)[2],(z)[3],
				(d)[1],(d)[2],(d)[3],
				(g)[0],(g)[1],(g)[2],(g)[3]);
			BANDPASS2N4P((ein),(aus),
				(z)[4],(z)[5],(z)[6],(z)[7],
				(d)[4],(d)[5],(d)[6],
				(g)[4],(g)[5],(g)[6],(g)[7]);
		}
		break;
	case 3:	
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER1P((ein),(aus),(z)[0],(g)[0]);
			FILTER1P((ein),(aus),(z)[1],(g)[0]);
			FILTER1P((ein),(aus),(z)[2],(g)[0]);
			FILTER1P((ein),(aus),(z)[3],(g)[0]);
			FILTER1P((ein),(aus),(z)[4],(g)[0]);
			FILTER1P((ein),(aus),(z)[5],(g)[0]);
			FILTER1P((ein),(aus),(z)[6],(g)[0]);
			FILTER1P((ein),(aus),(z)[7],(g)[0]);
		} else {
			BANDPASS1N2P((ein),(aus),(z)[0],(z)[1],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[2],(z)[3],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[4],(z)[5],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[6],(z)[7],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[8],(z)[9],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[10],(z)[11],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[12],(z)[13],
				(d)[1],(d)[2],(g)[0],(g)[1]);
			BANDPASS1N2P((ein),(aus),(z)[14],(z)[15],
				(d)[1],(d)[2],(g)[0],(g)[1]);
		}
		break;
	case 5:
	case 16:
	case 10:
		AMPLIFIER((ein),(aus),(d)[0]);
		if ((code) > 0) {
			FILTER2P((ein),(aus),(z)[0],(z)[1],(g)[0],(g)[1]);
			FILTER2P((ein),(aus),(z)[2],(z)[3],(g)[2],(g)[3]);
			FILTER2P((ein),(aus),(z)[4],(z)[5],(g)[4],(g)[5]);
			FILTER2P((ein),(aus),(z)[6],(z)[7],(g)[6],(g)[7]);
		} else {
			BANDPASS2N4P((ein),(aus),
				(z)[0],(z)[1],(z)[2],(z)[3],
				(d)[1],(d)[2],(d)[3],
				(g)[0],(g)[1],(g)[2],(g)[3]);
			BANDPASS2N4P((ein),(aus),
				(z)[4],(z)[5],(z)[6],(z)[7],
				(d)[4],(d)[5],(d)[6],
				(g)[4],(g)[5],(g)[6],(g)[7]);
			BANDPASS2N4P((ein),(aus),
				(z)[8],(z)[9],(z)[10],(z)[11],
				(d)[7],(d)[8],(d)[9],
				(g)[8],(g)[9],(g)[10],(g)[11]);	
			BANDPASS2N4P((ein),(aus),
				(z)[12],(z)[13],(z)[14],(z)[15],
				(d)[10],(d)[11],(d)[12],
				(g)[12],(g)[13],(g)[14],(g)[15]);
		}
		break;
	}
	*ausp = aus;
}
/*******Dateiende********************************************************/

