/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. It
 *   implements research work carried out by the Acoustical Communications
 *   Group at the Technical University of Munich. All rights reserved. FOR
 *   LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	TVF.C							*/
/*  Version  :	19.10.98/8						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:	Modul zur zeitvarianten Filterung mit der FTT.		*/
/*									*/
/*		Die Kernfunktionen sind splt_tvf, modu_tvf, synt_tvf.	*/
/*		splt_tvf nimmt einen Abtastwert des Eingangssignals	*/
/*		und verteilt ihn, multipliziert mit einem Kanal-	*/
/*		individuellen Verstaerkungsfaktor GAIN auf die 		*/
/*		Eingaenge der komplexen Modulatoren der einzelnen	*/
/*		Kanaele. Diese Eingaenge koennen im Statustabellenfeld 	*/
/*		IN gelesen und modifiziert werden. modu_tvf realisiert 	*/
/*		die komplexe Modulaton mit der Kanalfrequenz. Die Ergeb-*/
/*		nisse in Real- und Imaginaerteil koennen danach extern	*/
/*		im Statustabellenfeld RS bzw IS gelesen und modifiziert */
/*		werden.	synt_tvf synthetisiert die Frequenzkanaele	*/
/*		durch Tiefpassfilterung und komplexe Rueckmodulation	*/
/*		mit der Kanalfrequenz wieder zu einem Zeitsignal.	*/
/*									*/
/*  History  :  9.2.91/1, Modifikation aus TVF.C/H 14.12.90/3		*/
/*		20.1.94/2, Einbau von gfp->idxmod zum Ueberspringen	*/
/*		von Frequenzindices in den Kernfunktionen.		*/
/*		2.2.94/3 Stuetzstellenfrequenzabstand DF mitberechnet.	*/
/*		10.2.94/4 Synthese-Skalierung um Faktor 2 herabkorrig.	*/
/*		spmo_tvf() eingefuehrt, Analyse/Synthese-TC-Impulsgew.	*/
/*		Schoneheits-Korrekturen.				*/
/*		2.7.98/5 parm->ATB aus 20.1.94/2 wieder eingefuegt 	*/
/*		9.8.98/6 PI nur def wenn undef 				*/
/*		28.8.98/7 ERB- und SPINC-Skala 				*/
/*		19.10.98/8 Private Includes von TVF.H -> TVF.C,		*/
/*		    dabei System- vor Paket-Includes			*/
/*									*/
/************************************************************************/

#define	TVF_SOURCE			/* Source-Identifikation fuer...*/
#include <math.h>
#include <stdio.h>
#include "tvf.h"			/* Import/Export-Datei		*/

#ifndef PI
#define PI		3.141592654	/* PI				*/
#endif

/*----------------------------------------------------------------------*/
/*  cbw_at_freq		errechnet Frequenzgruppenbreite bei einer	*/
/*			Frequenz f. Einheiten jeweils Hz.		*/
/*----------------------------------------------------------------------*/
static double cbw_at_freq(f)
double f;
{	/* nach Zwicker, Terhardt JASA 68(5), 1523-1525, 1980 */
	return(25. + 75. * exp(.69 * log(1 + 1.4 * f * f * 1E-6)));
}
/*----------------------------------------------------------------------*/
/*  erb_at_freq		errechnet ERB bei einer Frequenz f.		*/
/*			Einheiten jeweils Hz.				*/
/*----------------------------------------------------------------------*/
static double erb_at_freq(f)
double f;
{
	/* Moore, B.C.J. & Glasberg, B.R., "Suggested formulae for
	 * calculating auditory-filter bandwith and excitation patterns",
	 * JASA 74(3), pp. 750-753, 1983
	 */
	return(6.23*(f*f*1E-6)+93.39*(f*1E-3)+28.52);
}
/*----------------------------------------------------------------------*/
/* spinc_at_freq	errechnet spinc bei einer Frequenz f.		*/
/*			Einheiten jeweils Hz.				*/
/*----------------------------------------------------------------------*/
static double spinc_at_freq(f)
double f;
{
	/* Terhardt, E. "The SPINC function for scaling of frequencies
	 * in auditory models", Acustica 77, 40-42, 1992
	 */
	return(1+(f/1414)*(f/1414));
}
/*----------------------------------------------------------------------*/
/*  PARM_TVF	errechnet die Stuetzfrequenzen fuer das TVF-Spektrum,	*/
/*		sowie die Transformationsparameter-Tabelle fuer die	*/
/*		Funktion do_tvf(). Die dazu benutzete Funktion		*/
/*		filterparms() wird per arithm.h importiert.		*/
/*									*/
/*    ->|	gfp->nfi	Anzahl Stuetzfrequenzen			*/
/*    		     fa		Abtastfrequenz Zeitfunktion in Hz.	*/
/*    		     f0		unterste Stuetzwertfrequenz in Hz.	*/
/*    		     df		Abstand A.-Freq. in folg. Einheit:	*/
/*		     unitdf	HZ,PROZ(%-Aufschlag z.vorig.Wert),..	*/
/*		     b		3dB-Synthesebandbreite in folg. Einh.:	*/
/*		     unitb	HZ, PROZ, BARK, ERB, SPINC		*/
/*		     tcb	Zeitkontur-Analysefilter-B3dB		*/
/*		     tcunitb	und Einheit				*/
/*		     tcfiltcode	Filtertyp Zeitkonturanalyse		*/
/*		     tgcomp	Kompensationsgrad (0..1) der max.	*/
/*		     		Laufzeit Synthese-TP			*/
/*		     sfiltcode	Filtertyp Synthese			*/
/*				Kanal-B3dB / Kanalabtastfrequenz 	*/
/*		gfs->parmtab	Tabelle Parameter an einer St.-Freq.	*/
/*		     maxdly	Maximale zeitl. Verzoeg. in Abt.W zw.	*/
/*				tiefster und hoechster St.-frequenz.	*/
/*	|->	(parmtab)	Inhalt wird initialisiert		*/
/*----------------------------------------------------------------------*/
static parm_tvf(gfp, gfs)
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	double tcb, tg, delay, ph = 0, fg,
		f = gfp->f0, ta = 1./gfp->fa;
	int fsflag = 0;
	long fi;
	register AFREQPARM *parm;
					/* Schleife ueber alle Analyse-	*/
					/*   frequenzen, aktuell: f	*/
	for (fi = 0; fi < gfp->nfi; fi++) {
		parm = gfs->parmtab+fi;	/* leichterer Zugriff		*/
					/* Frequenzgruppenbreite	*/
		parm->FREQ = f;		/* Stuetzfrequenz		*/
		parm->CBW = cbw_at_freq(f);
					/* Modulatorphasenfortschritt	*/
		parm->MDP = (SININP)(f * ta * 0x10000);

					/* fuer Synthesefilter		*/
		switch (gfp->unitb) {	/* abs.Bandbreite aus b,entspr.	*/
			default:	/*   kplxer BB des aequiv. TPs	*/
			case HZ:    parm->B = gfp->b; break;
			case PROZ:  parm->B = gfp->b * f / 100; break;
			case BARK:  parm->B = gfp->b * parm->CBW; break;
			case ERB:   parm->B = gfp->b * erb_at_freq(f); break;
			case SPINC: parm->B = gfp->b * spinc_at_freq(f); break;
		}
					/* Bandbreitenmass bzgl. 1.Tbd.	*/
		parm->ATB  = 10*log10(parm->B/(parm-fi)->B);
					/* Begrenzung kplxe BB auf	*/
					/*   halbe Abtastfrequenz	*/
		if (parm->B > 1./(ta * 2.))
		    parm->B = 1./(ta * 2.);

					/* fuer TC-Analysefilter	*/
		switch (gfp->tcunitb) {	/* abs.Bandbreite aus b,entspr.	*/
			default:	/*   kplxer BB des aequiv. TPs	*/
			case HZ:    tcb = gfp->tcb; break;
			case PROZ:  tcb = gfp->tcb * f / 100; break;
			case BARK:  tcb = gfp->tcb * parm->CBW; break;
			case ERB:   tcb = gfp->tcb * erb_at_freq(f); break;
			case SPINC: tcb = gfp->tcb * spinc_at_freq(f); break;
		}
		if (tcb > 1./(ta * 2.))	/* Begrenzung kplxe BB auf	*/
		    tcb = 1./(ta * 2.);
					/* Hoehe Impulsantwort von	*/
					/* TC-Analysefilter auf	Impuls-	*/
					/* flaeche eins.		*/
		parm->TCAIM = hnmax(gfp->tcfiltcode) * tcb * PI;

					/* Koeffizienten TP-Filter ber.	*/
					/* Rueckgabewert: Laufzeit in	*/
					/*   Filtermitte (f=0) in sec	*/
		tg = filterparms(gfp->sfiltcode,gfp->fa,parm->B/2.,
			f,ph,parm->AD,parm->AG);

		if(fi == 0) 		/* Filterbandbr. faellt nicht	*/
		    gfs->tgmax = tg;	/* mit steigendem Freq.Index,	*/
		    			/* also kann laengste Laufzeit	*/ 
					/* bei fi==0 angenommen werden.	*/

		delay = 		/* Gruppenlaufzeitkorrektur	*/
		    gfp->tgcomp * 	/*   Verzoegerung in Abt.Wtn.	*/
			(gfs->tgmax - tg) / ta;
					/* Anm.: da Grp.-Lz. im Durch-	*/
					/* lassber. mehr od. weniger	*/
					/* schwankt, gilt Korrektur n.	*/
					/* streng u. kann ueber tgcomp	*/
					/* "getrimmt" werden.		*/

					/* Filterlaufzeiten in Abtastw.	*/
					/* sowie Feinaufloesung in	*/
					/* Bruchteilen eines Abtastw.	*/
					/* Analysefilter		*/
		parm->ATG = (int)(tg / ta);
		parm->ATF = tg / ta - parm->ATG;
					/* Verzoegerung fuer Laufzeit-	*/
					/* kompensation in Abtastwerten	*/
					/* bzw. Bruchteilen		*/
		parm->DLY = (int)delay;
		parm->DLF = delay - parm->DLY;

					/* Allpasskoeffizient r		*/
		parm->DG  = (1. - parm->DLF) / (1. + parm->DLF) - 1E-6;
					/* Abstand TC-Analysefrequenz	*/
		switch (gfp->tcunitdf) {
		    default:
		    case HZ:   parm->TCDF = gfp->tcdf; break;
		    case PROZ: parm->TCDF = gfp->tcdf * f / 100. ; break;
		    case BARK: parm->TCDF = gfp->tcdf * parm->CBW; break;
		    case ERB:   parm->TCDF = gfp->tcdf * erb_at_freq(f); break;
		    case SPINC: parm->TCDF = gfp->tcdf * spinc_at_freq(f); break;
		}	
		switch (gfp->unitdf) {	/* Abstand Synthesefrequenz	*/
		    default:
		    case HZ:   parm->DF = gfp->df; break;
		    case PROZ: parm->DF = gfp->df * f / 100. ; break;
		    case BARK: parm->DF = gfp->df * parm->CBW; break;
		    case ERB:   parm->DF = gfp->df * erb_at_freq(f); break;
		    case SPINC: parm->DF = gfp->df * spinc_at_freq(f); break;
		}	
					/* Gewicht (Flaeche)		*/
					/* fuer in Synthesefilter	*/
					/* einzuspeisenden Impuls,	*/
					/* Damit Impulsflache eins	*/
					/* am tvf-Ausgang herauskommt	*/
		parm->TCSIW = gfp->idxmod * parm->DF /
			(hnmax(gfp->sfiltcode) * parm->B * PI);
		if (isp1(gfp->sfiltcode))
					/* Wenn Synthesefilter nur 1Pol	*/
					/* ist doppeltes Gewicht noetig	*/
					/* weil h(n=0) = 0.5;		*/
			parm->TCSIW *= 2.0;

		f += parm->DF;		/* naechste A.-Freq. berech. 	*/
					/* Maximal noetige Laufzeitk.	*/
	}				/* an hoechster A-Freq.		*/
	gfs->maxdly = (gfs->parmtab+(long)(gfp->nfi-1))->DLY;
	gfs->maxdlf = (gfs->parmtab+(long)(gfp->nfi-1))->DLF;
					/* Laufzeitfeinkorrektur aus/an	*/
	gfs->allpass = (gfp->tgcomp == 0.) ? 0 : 1;
}
/*----------------------------------------------------------------------*/
/*  RES_STAT	setzt die Zustandsspeicher aller Analysefreq. zurueck.	*/
/*									*/
/*    ->|	gfp->nfi	Anzahl Analysefrequenzen		*/
/*		gfs->stattab	Tabelle der Zustandspeicher		*/
/* 	|->	gfs->syffdesc	Fifo-Zustand				*/
/*		(stattab)	Inhalt stattab				*/
/*----------------------------------------------------------------------*/
static res_stat(gfp, gfs)
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	long fi, zi;
	AFREQSTAT *stat;
	AFREQPARM *parm;
					/* Zustandsspeicher loeschen	*/
	for (fi = 0; fi < gfp->nfi; fi++) {
		parm = gfs->parmtab + fi;
		stat = gfs->stattab + fi;
		for (zi = 0; zi < MAXZ; zi++) {
			stat->RSZ[zi] = 0;
			stat->ISZ[zi] = 0;
		}
		stat->IN = 0;
		stat->RS = 0;
		stat->IS = 0;
		stat->RSD = 0;
		stat->ISD = 0;
		stat->MPH = 0;
		stat->GAIN= 1.;		/* Verstaerkung eins		*/
	}
	clearfifo(&(gfs->syffdesc));	/* Synthese-Fifo mit 0 fuellen	*/
	fillfifo(&(gfs->syffdesc),(double)(0));
}
/*----------------------------------------------------------------------*/
/*  SPLT_TVF	Verteilung des Eingangsabtastwertes *tfunptr auf die 	*/
/*		Modulatoreingaenge im Statustabellenfeld IN,		*/
/*		nach Multiplikation mit Statustabellenfeld GAIN.	*/
/*----------------------------------------------------------------------*/
splt_tvf(tfunptr, gfp, gfs)
TIMSIG *tfunptr;
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	long fi;
	register AFREQSTAT *stat;
	int idxmod = gfp->idxmod;

	stat = gfs->stattab;		 /* Schleife ueber f-Stuetzw...	*/
	for (fi = 0; fi < gfp->nfi; fi+=idxmod) {
		stat[fi].IN = *tfunptr * stat[fi].GAIN / ARITHM_ZH15M1;
	}
}
/*----------------------------------------------------------------------*/
/*  MODU_TVF	Eingangsmodulatorteil der zeitvarianten Filterung.	*/
/*		Die eigenlichen Rechenfunktionen SINUSCOSINUS,		*/
/*		MODULATOR erfolgen ueber Makros, die in der 		*/
/*		include-Datei arithm.h festgelegt sind.			*/
/*----------------------------------------------------------------------*/
modu_tvf(gfp, gfs)
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	long fi;
	SININP phase;
	SINOUT si, co;
	register AFREQSTAT *stat;
	register AFREQPARM *parm;
	int idxmod = gfp->idxmod;

					/* Schleife ueber f-Stuetzw...	*/
	for (fi = 0; fi < gfp->nfi; fi+=idxmod) {
		stat = gfs->stattab+fi;	/* Beschleunigung Zugriff	*/
		parm = gfs->parmtab+fi;
					/* Modulatorphase weiterschalt.	*/
		phase = (stat->MPH -= parm->MDP);
					/* sin()/cos berechnen		*/
		SINUSCOSINUS(phase, si, co);
					/* cos/sin-Modulator 		*/
		stat->RS =(stat->IN / ARITHM_ZH15M1) * co;
		stat->IS =(stat->IN / ARITHM_ZH15M1) * si;
	}
}
/*----------------------------------------------------------------------*/
/*  SPMO_TVF	Ersetzt splt_tvf und modu_tvf, wenn nur Synthese, aber	*/
/*		keine Zeitvariante Filterung benoetigt wird.		*/
/*----------------------------------------------------------------------*/
spmo_tvf(gfp, gfs)
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	int fi;
	int idxmod = gfp->idxmod;
	register AFREQSTAT *stat;
					/* Schleife ueber f-Stuetzw...	*/
	for (fi = 0; fi < gfp->nfi; fi+=idxmod) {
		stat = gfs->stattab+fi;	/* Beschleunigung Zugriff	*/
					/* Modulatorphase weiterschalt.	*/
		stat->MPH -= gfs->parmtab[fi].MDP;
		stat->RS = 0;
		stat->IS = 0;
	}
}
/*----------------------------------------------------------------------*/
/*  SYNT_TVF	Syntheseteil mit Rekonstruktionsfilter.			*/
/*		Die eigenlichen Rechenfunktionen SINUSCOSINUS,		*/
/*		MODULATOR, ADDSQUARE und FILTERING erfolgen ueber	*/
/*		Makros, die in der include-Datei arithm.h festgelegt	*/
/*		sind. 							*/
/*----------------------------------------------------------------------*/
synt_tvf(gfp, gfs, tfunptr)
TIMSIG *tfunptr;
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	long fi;
	SINOUT si, co;
	FIFODESCR *fdp;
	register SININP ph;
	int fcode = gfp->sfiltcode;
	double re_ein, re_aus, im_ein, im_aus, abtwversatz, scale;
	FIFOELEM ffelem;

	register AFREQSTAT *stat;
	register AFREQPARM *parm;
	int idxmod = gfp->idxmod;

	fdp = &(gfs->syffdesc);		/* Fifo-Descriptor		*/
	scale = gfp->scale;
					/* Zeitpunkt max. Fensteroeffn.	*/
	abtwversatz = gfp->syntadj * gfp->tgcomp * gfs->tgmax * gfp->fa;

	for (fi = 0; fi < gfp->nfi; fi+=idxmod) {
		stat = gfs->stattab+fi;	/* Beschleunigung Zugriff	*/
		parm = gfs->parmtab+fi;

		ph = stat->MPH;
		ph -= (unsigned WORD)	/* wg. Fifo nach Modul.		*/
			((long)(parm->MDP) * parm->DLY);
		ph += (unsigned WORD) (parm->MDP * abtwversatz);
		SINUSCOSINUS(ph, si, co);

		re_ein = stat->RS;
		im_ein = stat->IS;
		FILTERING(fcode,re_ein,
			stat->RSZ,parm->AD,parm->AG,re_aus);
		FILTERING(fcode,im_ein,
			stat->ISZ,parm->AD,parm->AG,im_aus);
					/* Laufzeitfeinabgleich		*/
		if (gfs->allpass) {
			re_ein   = re_aus;
			re_aus   = stat->RSD + re_ein * parm->DG;
			stat->RSD= re_ein - re_aus * parm->DG;
			im_ein   = im_aus;
			im_aus   = stat->ISD + im_ein * parm->DG;
			stat->ISD= im_ein - im_aus * parm->DG;
		}
					/* Beitrag ensprechend Verzoeg.	*/
					/* in Fifo addieren		*/
		addfifo(fdp, -1 -(gfs->maxdly - parm->DLY), 	
			(FIFOELEM)(re_aus * co + im_aus * si));
	}
	outelemfifo(fdp, &ffelem);	/* vorderstes Fifoelement raus	*/
	tfunptr[0] = (WORD)(scale * ffelem);
	eleminfifo(fdp, (double)0);	/* in Fifo null nachschieben	*/
}
/*----------------------------------------------------------------------*/
/*  INFO_TVF	holt Eintraege aus der Filtertypentabelle. Dabei ist	*/
/*		Index die Nummer des Eintrages (ab 0), ueber den eine	*/
/*		Tabellenzeile geholt wird. Der erste Zeileneintrag ist	*/
/*		int-Zahl, die Filterkennummer. Der zweite und dritte	*/
/*		sind Strings, die den Kurz- bzw. den ausf. Namen ent-	*/
/*		halten. Zurueck kommen die besetzen Zeiger. Zusaetzlich	*/
/*		wird ein Zeiger auf einen Kommentar zur Rechenreali-	*/
/*		sierung besetzt. (return) ungleich null, wenn Index zu 	*/
/*		hoch.							*/
/*----------------------------------------------------------------------*/
int info_tvf(index, codep, abrp, namep, commentp)
int index, *codep;
char **abrp, **namep, **commentp;
{
	int i = 0;

	*commentp = ARITHM_COMMENT;
	while((*codep = g_ftyptab[i].FTYPCODE) != 0) {
		if (index == i) {
			*abrp  = g_ftyptab[i].FTYPABR;
			*namep = g_ftyptab[i].FTYPNAME;
			return(0);
		}
		i++;
	}
	return(-1);
}
/*----------------------------------------------------------------------*/
/*  INIT_TVF 	Diese Funktion waehlt dyn. Speicher der zwei Tabellen	*/
/*		und des Fifos an, berechnet per parm_tvf() die Trans-	*/
/*		formationstabelle und setzt den Status der Analysefre-	*/
/*		quenzen ueber res_stat() zurueck.			*/
/*		(return) ungleich 0 bei Misserfolg: Speichermangel.	*/
/*		(Parameter siehe parm_tvf() und res_stat())		*/
/*----------------------------------------------------------------------*/
int init_tvf(gfp, gfs)
GTVFPARMS *gfp;
GTVFSTAT *gfs;
{
	if (gfp->idxmod == 0)
		gfp->idxmod = 1;
	if ((gfs->parmtab = ((AFREQPARM*)lmalloc(sizeof(AFREQPARM) * 
			(long)gfp->nfi))) == 0L)
		return(-1);
	if ((gfs->stattab = ((AFREQSTAT*)lmalloc(sizeof(AFREQSTAT) * 
			(long)gfp->nfi))) == 0L) {
		free((char*)(gfs->parmtab));
		return(-1);
	}
	parm_tvf(gfp, gfs);
	if (createfifo(&(gfs->syffdesc), gfs->maxdly + 2)) {
		free((char*)(gfs->parmtab));
		free((char*)(gfs->stattab));
		return(-1);
	}
	res_stat(gfp, gfs);
	return(0);
}
/*----------------------------------------------------------------------*/
/*  DISPOSE_TVF	gibt den fuer die TVF angewaehlten Speicher wieder frei	*/
/*----------------------------------------------------------------------*/
dispose_tvf(gfs)
GTVFSTAT *gfs;
{
	free((char*)(gfs->parmtab));
	free((char*)(gfs->stattab));
	disposefifo(&(gfs->syffdesc));
}
/*******Dateiende********************************************************/

