/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. It
 *   implements research work carried out by the Acoustical Communications
 *   Group at the Technical University of Munich. All rights reserved. FOR
 *   LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	TVF.H - Import/Export-Datei zu -> TVF.C			*/
/*  Version  :	19.10.98/8						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:	Zeitvariante Filterung mit Fourier-t-Transformation. 	*/
/*									*/
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -	*/
/*									*/
/*  Folgenden Funktionen sind verfuegbar:				*/
/*									*/
/*	info_tvf()	Informationen Filtertypen: Code, Kuerzel, Name,	*/
/*			Kommentar Rechnungsrealisierung			*/
/*	init_tvf()	TVF anmelden					*/
/*	splt_tvf()	TVF Auffaecherungsteil und Verstaer.-steuerrng.	*/
/*	modu_tvf()	TVF Analyseteil, nur Modulatoren 		*/
/*	synt_tvf()	TVF Syntheseteil mit Rekonstruktionsfilter 	*/
/*	dispose_tvf()	TVF abmelden					*/
/*									*/
/************************************************************************/

#ifndef TVF_INCLUDED			/* Schutz vor mehrfach-include	*/
#define TVF_INCLUDED

/*----------------------------------------------------------------------*/
/*  Import - wird auch Export:						*/
/*----------------------------------------------------------------------*/

#include "portab.h"	/* Portabilitaetsdatei: WORD,lmalloc()...	*/
#include "doubfifo.h"	/* createfifo(),clearfifo(),disposefifo()	*/
			/*   inelemfifo(),outelemfifo(),readfifo()	*/
#include "filtflp.h"	/* filterparms(),fitering(),MAXG/MAXZ/MAX/D	*/
#include "sintab.h"	/* SINUSCOSINUS(),SINUS(),COSINUS()		*/
#include "hnmax.h"	/* Ergaenzung zu filtflp, Maxima norm. IP-Antw.	*/
					
/*----------------------------------------------------------------------*/
/*  Export :								*/
/*----------------------------------------------------------------------*/

extern int info_tvf();		/* Information TVF-Filtertypen		*/
extern int init_tvf();		/* TVF einrichten			*/
extern splt_tvf();		/* TVF Auffaecherungsteil 		*/
extern modu_tvf();		/* TVF Analyseteil 			*/
extern spmo_tvf();		/* TVF Auffaecherung+Analyse m. input=0	*/
extern synt_tvf();		/* TVF Syntheseteil mit Rekon.		*/
extern dispose_tvf();		/* TVF abmelden				*/

				/* Max. 40 Zeichen			*/
#define ARITHM_COMMENT	"Floatingpoint, sin()/cos() tabelliert"

				/* Typdefinitionen und verwendete Wer-	*/
				/* tebereiche (Calibrierung s.u.):	*/
#define TIMSIG	WORD		/* Zeitsignal		+/- 2**15-1	*/
#define	SININP  unsigned WORD	/* sin()/cos()-Argument	0...2**16-1	*/
#define SINOUT	WORD		/* sin()/cos()-Wert	+/- 2**15-1	*/
#define	FILCOE	double		/* Filterkoeffizenten			*/
#define	FILTER	double		/* Z,In,Out von Filter			*/
#define LEVSP	double		/* Pegel (mB = 0.01dB)	+/- 2**15-1(mB)	*/


#define ARITHM_ZPI	6.283185308	/* 2*PI				*/
#define	ARITHM_ZH15M1ZQ	1.073676289e9	/* ((2**15)-1)**2		*/
#define ARITHM_ZH15M1	32767.		/* 2**15-1			*/
#define ARITHM_ZH16	65536.		/* 2**16			*/


#define FILTERING(code,in,z,d,g,aus)					\
				filtering((code),(in),(z),(d),(g),&(aus))
#define MODULATOR(sig,car,out)	(out)=(((sig) / ARITHM_ZH15M1ZQ) * (car));
#define ADDSQUARE(in1,in2,out)	(out)=((in1)*(in1) + (in2)*(in2));
#define LEVEL(in,offset,out)						\
{									\
 (out)=((in) > 1e-30 ? (10.*log10(in) + 90.) : -210.) + (offset);	\
 (out)=((out) > 0. ? (out) * 100 : 0.);					\
}
					
					/* Parameter fuer eine An.-Freq.*/
					/* (f. Anwender normw.n.relev.)	*/
typedef	struct {			/*   Filterkoeffizienten	*/
	FILCOE	AG[MAXG],		/*     Analysefilter		*/
		AD[MAXD],
		DG;			/*     Laufzeitfeinkorrektur	*/
	SININP	MDP;			/*   Phasenfortschritt Modulat.	*/
	int	DLY,			/*   Verz.i.Abtw.z.Laufztkomp.	*/
		ATG;			/*   Filter-Lz. in Abtwn	*/
	double  DLF,			/*   Feinaufloesung der Lz. in	*/
		ATF,			/*   Bruchteilen eines Abtw.	*/
		FREQ,			/*   Synthesefrequenz in Hz	*/
		CBW,			/*   Frequenzgruppenbreite/Hz	*/
		B,			/*   3dB-Synth.-filterbandb./Hz	*/
		ATB,			/*   BB-Mass in dB bzgl. 1. Tbd.*/
		DF,			/*   Stuetzstellenfrequenzabst.	*/
		TCDF,			/*   TC-Analysefrequenzabst./Hz	*/
		TCAIM,			/*   TC-Analyse-Impulsantw.-	*/
					/*     Max.f.Impulsflaeche eins	*/
		TCSIW;			/*   Impulsgewicht (Flaeche)	*/
					/*     an Synthesefiltereingang	*/
					/*     fuer TVF-Summenausgangs-	*/
					/*     Impulsgew. gleich eins	*/
} AFREQPARM;

					/* Zust. an einer Analysefreq.:	*/
typedef	struct {			/*   Zustandsspeicher Filter 	*/
	FILTER	RSZ[MAXZ],		/*     Synthese Realteilzweig	*/
		ISZ[MAXZ],		/*     Synthese Imaginaerteilzw.*/
		RSD,			/*     Lz.-Feinkorr. Realteil	*/
		ISD,			/*     Lz.-Feinkorr. Imaginaert.*/
		GAIN,			/*   Verstaerkungssteuerung	*/
		IN,			/*   Eingangsmodulatoren,	*/
		RS,			/*   	Eingang und Ausgaenge	*/
		IS;			/*      in Real- und Imag.-T.	*/
	SININP	MPH;			/*   Zustandsspeicher Modulator	*/
} AFREQSTAT;


typedef	struct {			/* Allgemeine TVF-Parameter	*/
	double		fa,		/*   Abtastfrequenz in Hz	*/
			f0,		/*   tiefste Syn.-Freq in Hz	*/
			df,		/*   Abstand S.-Freq in unitdf	*/
			tcdf,		/*   Abst. TC-A.-Freq tcunitdf	*/
			b,		/*   3dB-S.-Bandbreite in unitb	*/
			tcb,		/*   TC-Analyseb. in tcunitb	*/
			offset,		/*   Pegeloffset 		*/
			tgcomp,		/*   Kompensation Gruppenlaufzt.*/
			scale,		/*   Syntheseskalierungsfaktor	*/
			syntadj;	/*   Synthesezeitpunktjustage	*/
	int		nfi,		/*   Anz. Synthesefrequenzen	*/
			idxmod,		/*   Indexmodulo der AFs	*/
			unitdf,		/*   Einheitenkennung fuer df,b */
			unitb,		/*    in HZ,PROZ,BARK,ERB,SPINC	*/
			tcunitdf,	/*     und Zeitkonturpunktabst.	*/
			tcunitb,	/*     Zeitkontur-Analyse-BB	*/
			tcfiltcode,	/*   Code TC-Analysefiltertyp	*/
			sfiltcode;	/*   Code Synthesefiltertyp	*/
} GTVFPARMS;

typedef struct {			/* Allgemeiner TVF-Status	*/
	AFREQPARM 	*parmtab;	/*   Tabelle St.-Freq.Parameter	*/
	AFREQSTAT	*stattab;	/*   Tabelle St.-Freq. Zustand	*/
	FIFODESCR	syffdesc;	/*   Fifo-Descriptor		*/
	int		maxdly;		/*   Maxim. Verzoeg. in Abtw.	*/
	double		maxdlf;		/*   plus in Bruchteilen v. Abt.*/
	int		allpass;	/*   Flag Laufzeitfeinabgleich	*/
	double		tgmax;		/*   Laufzeit in sec. fuer	*/
					/*   tiefstes Synthesefilter in */
					/*   Filtermitte (kplx.)	*/
} GTVFSTAT;


#define HZ		0		/* moegliche Einheiten fuer 	*/
#define PROZ		1		/*   unitdf, unitb		*/
#define BARK		2
#define ERB		3
#define SPINC		4

#endif					/* TVF_INCLUDE			*/
/*******Dateiende********************************************************/

